package org.greenstone.gsdl3.gs3build.doctypes;

import java.io.*;

import java.net.URL;

import javax.xml.parsers.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.greenstone.gsdl3.gs3build.metadata.*;

public class METSDocument extends AbstractDocument
{
	public static final String METS_DOCUMENT_TYPE = "METS";

	public METSDocument(URL url)
	{	super(url);

		if (url.toString().startsWith("file://"))
		{	this._parseFile(new File(url.toString().substring(7)));
		}
	}

	private void _parseFile(File file)
	{
	  try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse(file);

			// TODO: get all the types in the tree

			// TODO: do a traverse, and thus cope with elements-within-elements if needsbe, but
			// this shouldn't happen except in a directly defined situation - actually doing the
			// parsing in part inside each node would work well provided one checked for a node
			// having already been done...
			System.out.println("Read");

			NodeList fileSecs = document.getElementsByTagName("mets:fileSec");

			for (int g = 0; g < fileSecs.getLength(); g ++) {
//				Schema schema = new Schema(schemas.item(s));
				this._parseFileSec(fileSecs.item(g));
			}

			// Get document metadata sections
			NodeList dmdSecs = document.getElementsByTagName("mets:dmdSec");

//				Schema schema = new Schema(schemas.item(s));
			this.metadata = METSDescriptiveSet.parseXML(fileSecs);
		}
		catch (FactoryConfigurationError e) {
			System.out.println(e);
		}
		catch (ParserConfigurationException ex) {
			System.out.println(ex);
		}
		catch (SAXException ex) {
			System.out.println(ex);
		}
		catch (IOException ex) {
			System.out.println(ex);
		}
	}

	private void _parseFileSec(Node fileSec)
	{	// this is in effect a group without a sense of 'self'...
		this._parseFileGroup((Element) fileSec, null);
	}

	private void _parseFileGroup(Element groupTag, METSFileGroup group)
	{	NodeList children = groupTag.getChildNodes();

		for (int c = 0; c < children.getLength(); c ++)
		{ if (children.item(c).getNodeType() != org.w3c.dom.Node.ELEMENT_NODE) {
				continue;
			}

			System.out.println(children.item(c));

			Element element = (Element) children.item(c);

			if (element.getNodeName().equals("mets:File"))
			{	if (group != null)
				{ METSFile file = METSFile.parseXML(element, group);
				}
				else
				{ // TODO: error
				}
			}
			else if (element.getNodeName().equals("mets:fileGrp"))
			{	// recurse
			  METSFileGroup childGroup = new METSFileGroup(element.getAttribute("ID"));

				this._parseFileGroup(element, childGroup);
				if (group != null)
				{ group.addGroup(childGroup);
				}
				else
				{ this.fileSet.addGroup(childGroup);
				}
			}
		}
	}

	public String getDocumentType()
	{	return METS_DOCUMENT_TYPE;
	}

	public String getDocumentText()
	{ // TODO: make this more than a dummy function!
		return null;
	}	
}