package org.greenstone.gsdl3.gs3build.util;

import java.net.*;
import java.io.*;
import java.applet.*;
import java.util.*;

/** 
 *  HTMLDOC: handles an individual HTML document
 */

public class HTMLDoc
{ String		 		content;
	String				title;
	URL						url;						 
	int						pos;
	int						status;
	HTMLBlockList blocklist;
	List          urls_out;
	List          urls_in;
	Object				note;
	HTMLLoader		loader;
	HTMLBlock			coded;

	/**
	 * Constructor which loads a given document; the ref indicates from where!
	 */
	public HTMLDoc(URL url, int ref)
	{ 
		URLConnection connect;
		int						length, i, c;
		byte					inbuffer[];
		InputStream		in;

		this.url 				= url;
		this.pos 				= 0;
		this.blocklist	= null;
		this.title			= null;
		this.urls_out		= null;
		this.urls_in		= null;
		this.note				= null;
		this.status			= 0;

		if (ref >= 0)
		{	int attempts;
		
			attempts = 0;
			loader = new HTMLLoader(url);
			loader.load();
			if (ref == 0)
			{ while(this.ready() == false)
				{	try
					{
						Thread.sleep(100);
					}
					catch (InterruptedException ex)
					{
					}
				}
			}
		}
		else
		{	this.content = "";
		}
	}

	public HTMLDoc(URL url)
	{	this(url, 0);
	}

	/**
	 * @return <code>true</code> when a document has finished loading.
	 */
	public synchronized boolean ready()
	{ if (this.loader == null)
		{	return true;
		}
	
		boolean loaded 	= loader.loaded();

		if (loaded == true)
		{ // reload if necessary
			if (this.loader.error == true)
			{	if (this.status < 3 && this.loader.notFound == false)
				{	this.status ++;
					this.loader = new HTMLLoader(this.url);
					loader.load();
					return false;
				}
				else
				{	this.status 	= 4;
					this.content	= null;
					this.loader		= null;
					return true;
				}
			}
			this.content 	= loader.HTML();
			this.loader		= null;
		}
		return loaded;
	}

	/**
	 * Create a document from a file
	 */
	public HTMLDoc(URL url, String file)
	{ FileInputStream _in;
		byte[] 					data = new byte[128];
		int							databytes;
		StringBuffer		content;

		try
		{	_in = new FileInputStream(file);
			if (_in != null)
			{ content 		= new StringBuffer();

				do
				{ databytes = _in.read(data);
					if (databytes > 0)
					{ content.append(new String(data, 0, databytes));
					}		
				} while (databytes >= 0);
				this.content		= content.toString();
			}
		}
		catch (IOException ex)
		{ this.content	= null;
			System.out.println(ex);
		}
		this.url				= url;
		this.pos				= 0;
		this.status			= 0;
		this.note				= null;
		this.blocklist	= null;
		this.title			= null;
		this.urls_out		= null;
		this.urls_in		= null;
	}

	/**
	 * Create a url-only document: the document body is not stored
	 */
	public HTMLDoc(String url)
	{ if (url != null)
		{	try
			{ this.url = new URL(url);
			}
			catch (MalformedURLException ex)
			{
				this.url		= null;
			}
		}
		else
		{	this.url 			= null;
		}
		this.pos				= 0;
		this.status			= 0;
		this.note				= null;
		this.content		= null;
		this.blocklist	= null;
		this.title			= null;
		this.urls_out		= null;
		this.urls_in		= null;
	}

	static public HTMLDoc fromString(String url, String string)
	{	HTMLDoc reply;

		reply 				= new HTMLDoc(url);
		reply.content	= string;
		return reply;
	}

	static public boolean validExtension(String url)
	{	String extension;

		// get file extension
		extension = URLString.extension(url);
		if (extension != null)
		{ extension = extension.toLowerCase();
		}

		if (extension == null ||
		    extension.indexOf("htm") >= 0 || 
		    extension.indexOf("asp") >= 0)
		{ return true;
		}
		return false;
	}

	static boolean validExtension(URL url)
	{	return validExtension(url.toString());
	}

	/**
	 * Uncache a single document
	 */
	public static HTMLDoc uncacheDoc(String filebase, String url)
	{	String 	pathname;
		String	filename;
		HTMLDoc	doc;

		pathname 	= URLString.pathName(url);
		if (pathname == null)
		{	return null;
		}

		filename	=	filebase + pathname.replace('/', File.pathSeparatorChar);
		doc	= new HTMLDoc(URLString.toURL(url), filename);

		return doc;
	}

	/**
	 * @return URL (ie. Java URL object) of the document
	 */
	public URL getUrl()
	{ return this.url;
	}

	// Report url to AppletContext
	public void show(AppletContext ac)
	{ URL url;

		url = this.getUrl();
		if (url != null)
		{ ac.showDocument(url);
		}
	}

	// Get a HTMLBlock list of the wanted text in an HTMLdoc
	public void getSelectContent(int level)
	{ HTMLBlockList list;

		list = new HTMLBlockList(this, level);
		this.blocklist = list;
	}

	/**
	 * return the selected content
	 */
	public HTMLBlockList selectContent()
	{ return this.blocklist;
	}

	/**
	 * get header blocks as selected content
	 */
	public HTMLBlockList headerBlocks()
	{	if (this.blocklist == null)
		{	this.getSelectContent(6);
		}
		return this.blocklist;
	}

	/**
	 * return a tagged-up select with an <A NAME=> at the beginning of each Hx
	 */
	public String taggedSelectContent()
	{ int heading, offset, pos;
		String reply;

		if (this.blocklist == null || 
				this.blocklist.size() == 0)
		{ return this.content;
		}

		reply = "";
		offset	= 0;
		for (heading = 0; heading < this.blocklist.size(); heading ++)
		{ reply = reply + 
							this.content.substring(offset, this.blocklist.tagPos(heading)) +
							"<A NAME=\"_L"+heading+"\"></A>";
			offset = this.blocklist.tagPos(heading);
		}
		reply = reply + this.content.substring(offset, this.content.length());
		return reply;
	}

	/**
	 * Return enumeration of headings in document
	 */
	public Enumeration headings()
	{ if (this.blocklist == null)
		{ this.getSelectContent(6);
		}
		return this.blocklist.elements();
	}

	/**
	 *	@return HTMLBlock of the child text underneath a heading
	 */
	public HTMLBlock headingChildText(int headref)
	{ HTMLBlock pageblock;
		HTMLBlock headblock;
		HTMLBlock	childblock;
		int	blockstart, blockend;

		pageblock = this.getCodedContent(); // get all the coded bits of the page
		headblock	= this.headerBlocks().tagBlock(headref);

		blockstart	= pageblock.find(headblock);
		blockstart	+= headblock.size();
		if (headref == this.headerBlocks().size() - 1)
		{ blockend	= pageblock.size();
		}
		else
		{ blockend	= pageblock.find(this.headerBlocks().tagBlock(headref+1));
		}
		childblock = pageblock.subBlock(blockstart, blockend);
		
		return childblock;
	}

	/**
	 * Add incoming url to list of incoming references
	 */
	public void addUrlIn(String url)
	{ if (this.urls_in == null)
		{ this.urls_in = new ArrayList();
		}
		this.urls_in.add(url);
	}

	/**
	 * @return a String of all incoming urls
	 */
	public String urlsInString()
	{ if (this.urls_in != null)
		{ return this.urls_in.toString();
		}
		return "";
	}

	/**
	 * @return an enumeration of all incoming urls as Strings
	 */
	public Iterator urlsIn()
	{ if (this.urls_in == null)
		{ return null;
		}
		return this.urls_in.iterator();
	}

	/**
	 * @return the <code>StringVector</code> of all incoming urls
	 */
	public List urlsInList()
	{	return this.urls_in;
	}

	/**
	 * @return <code>true</code> if the given url is linked to by this
	 *  document
	 */
	public boolean linkedTo(URL url)
	{	if (this.urls_out != null &&
				this.urls_out.indexOf(url.toString()) >= 0)
		{ return true;
		}
		return false;
	}

	/**
	 * See above: returns <code>true</code> if the document is linked to
	 * by this document.
	 */
	public boolean linkedTo(HTMLDoc doc)
	{ return linkedTo(doc.getUrl());
	}

	/**
	 * Add outgoing url to list of outgoing references
	 */
	public void addUrlOut(String url)
	{ if (this.urls_out == null)
		{ this.urls_out = new ArrayList();
		}
		this.urls_out.add(url);
	}

	/**
	 * @return a String of all outgoing urls
	 */
	public String urlsOutString()
	{ if (this.urls_out != null)
		{ return this.urls_out.toString();
		}
		return "";
	}

	/**
	 * @return the <code>StringVector</code> of all outgoing urls
	 */
	public List urlsOutList()
	{	return this.urls_out;
	}

	/**
	 * Return an enumeration of outbound urls
	 */
	public Iterator urlsOut()
	{ if (this.urls_out == null)
		{ return null;
		}
		return this.urls_out.iterator();
	}

	private static int urlListPos(List urlList, String url)
	{	for (int i = 0; i < urlList.size(); i ++)
		{	if (((String) urlList.get(i)).equals(url))
			{	return i;
			}
		}
		return -1;
	}

	// Rename internally the content of a list
	private void __renameUrl(List urlList, String oldUrl, String newUrl)
	{	if (urlList != null)
		{ int oldPos = urlListPos(urlList, oldUrl);

			if (oldPos >= 0)
			{ if (urlListPos(urlList, newUrl) >= 0)
				{ // delete oldUrl: now duplicate
					urlList.remove(oldPos);
				}
				else
				{ // rename oldUrl
					urlList.set(oldPos, newUrl);
				}
			}
		}
	}

	/**
	 * Rename inbound/outbound url references
	 */
	public void renameUrl(String oldUrl, String newUrl)
	{	this.__renameUrl(this.urls_in, oldUrl, newUrl);
	  this.__renameUrl(this.urls_out, oldUrl, newUrl);
	}

	/**
	 * Save document
	 */
	public boolean save(String filename, boolean select)
	{	FileOutputStream fout;
		PrintWriter pout;

		String flattened = URLString.flattenedUrl(filename);

		try
		{	HTMLDocAnchorList	anchorList	= new HTMLDocAnchorList(this);

			anchorList.flattenLinks();
			
			fout = new FileOutputStream(flattened);
			pout = new PrintWriter(fout);
			if (select == true)
			{	pout.print(this.taggedSelectContent());
			}
			else
			{	pout.print(this.content);
			}
			pout.close();
			fout.close();
		}
		catch (IOException ex)
		{	System.out.println(ex);
			return false;
		}
		return true;
	}

	/**
	 * Save document content
	 */
	public boolean save(String base)
	{ return this.save(base, false);
	}

	/**
	 * Save selected content
	 */
	public boolean saveSelect(String base)
	{	return this.save(base, true);
	}

	/**
	 * Get title of page
	 */
	public void getTitle()
	{ HTMLDocTitle doctitle;

		if (this.content != null)
		{	doctitle = new HTMLDocTitle(this);
			this.title	= doctitle.Title();
		}
	}

	/**
	 * @return: title of page
	 */
	public String titleString()
	{ if (this.title == null)
		{ this.getTitle();
		}
		return this.title;
	}

	/**
	 * @return a <code>String</code> of the url of the page
	 */
	public String urlString()
	{ return this.url.toString();
	}

	/**
	 * 	Replace the HTML text of the page with the provided HTML string
	 *	content.  This is an extremely dangerous call to use, as dependent
	 *	objects such as tag objects will be immediately rendered inaccurate
	 *	in terms of pointers held to the original text.
	 */
	protected void setContent(String newContent)
	{ this.content = newContent;
	}

	/**
	 * @return a <code>String</code> of the HTML text of the page
	 */
	public String getContent()
	{ return this.content;
	}

	/**
	 * @return a <code>HTMLBlock</code> of the page content
	 */
	public HTMLBlock getCodedContent()
	{	// if we haven't already got it, code the bugger up right now!
		if (this.coded == null)
		{ HTMLParser parser;
			String		next;
			HTMLBlock	coded;
			HTMLTag		tag;
			HTMLCText	text;

			if (this.content == null)
			{	return null;
			}

			parser 	= new HTMLParser(this);
			coded		= new HTMLBlock(0, this.content.length());

			parser.startParse();
			next	= parser.fullParse();
			while (next != null)
			{	if (HTMLTag.isTag(next))
				{ tag = new HTMLTag(next, parser.lastParse(), parser.atParse());
					coded.addTag(tag);
				}
				else
				{ text	= new HTMLCText(next, parser.lastParse(), parser.atParse());
					coded.addText(text);
				}
				next = parser.fullParse();
			}
			this.coded	= coded;
		}
		return this.coded;
	}
	
	/**
	 * @return <code>true</code> if this document is not errored from loading
	 * etc.
	 */
	public boolean valid()
	{	return (this.status < 4 && this.content != null && this.content.equals("null") == false);
	}

	/**
	 * HTMLString: returns content
	 */
	public String HTMLString()
	{ return this.content;
	}

	/**
	 * Return text of document as ascii text;
	 */
	 /*
	public String toString()
	{	StringBuffer	reply = new StringBuffer();
		int						parastart = 0;

		return reply.toString();
	}
	 */

	/**
	 * Set current note on this page
	 * @return old note object
	 */
	public Object setNote(Object note)
	{ Object oldnote;

		oldnote	= this.note;
		this.note = note;
		return oldnote;
	}

	/**
	 * @return current note <code>Object</code> on this page
	 */
	public Object getNote()
	{ return this.note;
	}
}